# Fortuna: A Collection of Random Value Generators for Python3
Fortuna's main goal is to provide a quick and easy way to build custom random-value generators for your data.

The core functionality of Fortuna is based on the RNG Storm engine. While Storm is a high quality random engine, Fortuna is not appropriate for cryptography of any kind. Fortuna is meant for games, data science, A.I. and experimental programming, not security.

Suggested Installation: `$ pip install Fortuna`

Precompiled for MacOS, installation on other platforms may require building from source.

Support this project: https://www.patreon.com/brokencode


### Table of Contents:
- Numeric Limits
- Project Terminology
- Random Generators:
    - Value Generators
        - `RandomValue(Collection) -> Callable -> Value`
        - `TruffleShuffle(Collection) -> Callable -> Value`
        - `QuantumMonty(Collection) -> Callable -> Value`
        - `CumulativeWeightedChoice(Table) -> Callable -> Value`
        - `RelativeWeightedChoice(Table) -> Callable -> Value`
        - `FlexCat(Matrix) -> Callable -> Value`
    - Integer Generators
        - `random_below(Integer) -> Integer`
        - `random_int(Integer, Integer) -> Integer`
        - `random_range(Integer, Integer, Integer) -> Integer`
        - `d(Integer) -> Integer`
        - `dice(Integer, Integer) -> Integer`
        - `plus_or_minus(Integer) -> Integer`
        - `plus_or_minus_linear(Integer) -> Integer`
        - `plus_or_minus_gauss(Integer) -> Integer`
    - Index Generators: 
        - ZeroCool Specification: `f(N) -> [0, N)` or `f(-N) -> [-N, 0)`
        - `random_index(Integer) -> Integer`
        - `front_gauss(Integer) -> Integer`
        - `middle_gauss(Integer) -> Integer`
        - `back_gauss(Integer) -> Integer`
        - `quantum_gauss(Integer) -> Integer`
        - `front_poisson(Integer) -> Integer`
        - `middle_poisson(Integer) -> Integer`
        - `back_poisson(Integer) -> Integer`
        - `quantum_poisson(Integer) -> Integer`
        - `front_geometric(Integer) -> Integer`
        - `middle_geometric(Integer) -> Integer`
        - `back_geometric(Integer) -> Integer`
        - `quantum_geometric(Integer) -> Integer`
        - `quantum_monty(Integer) -> Integer`
    - Float Generators
        - `canonical() -> Float`
        - `random_float(Float, Float) -> Float`
    - Boolean Generator
        - `percent_true(Float) -> Boolean`
    - Shuffle Algorithms, Inplace & Destructive
        - `shuffle(List) -> None`
        - `knuth(List) -> None`
        - `fisher_yates(List) -> None`
- Test Suite
    - `distribution_timer(Callable, *args, **kwargs) -> None`
    - `quick_test() -> None`
- Development Log
- Test Suite Output
- Legal Information


#### Numeric Limits:
- Integer: 64 bit signed integer.
    - Input & Output Range: `(-2**63, 2**63)` or approximately +/- 9.2 billion billion.
    - Minimum: -9223372036854775807
    - Maximum:  9223372036854775807
- Float: 64 bit floating point.
    - Minimum: -1.7976931348623157e+308
    - Maximum:  1.7976931348623157e+308
    - Epsilon Below Zero: -5e-324
    - Epsilon Above Zero:  5e-324


#### Project Terminology:
- Value: Almost anything in Python can be a Value.
    - Expressions, Generators, and F-strings should be wrapped in a lambda for dynamic evaluation.
- Callable: Any callable object, function, method or lambda.
- Collection: A group of Values.
    - List, Tuple, Set, etc... Any object that can be converted into a list via `list(some_object)`.
    - Comprehensions that produce a Collection also qualify.
    - Fortuna classes that wrap a Collection can wrap a Collection, Sequence or generator.
    - Fortuna functions that take a Collection as input will always require a Sequence.
- Sequence: An ordered Collection.
    - List, tuple or list comprehension.
    - A Sequence is an ordered Collection that can be indexed like a list, without conversion.
    - All Sequences are Collections but not all Collections are Sequences.
- Pair: Collection of two Values.
- Table: Collection of Pairs.
- Matrix: Dictionary of Collections.
- Inclusive Range.
    - `[1, 10] -> 1, 2, 3, 4, 5, 6, 7, 8, 9, 10`
- Exclusive Range.
    - `(0, 11) -> 1, 2, 3, 4, 5, 6, 7, 8, 9, 10`
- Partially Exclusive Range.
    - `[1, 11) -> 1, 2, 3, 4, 5, 6, 7, 8, 9, 10`
    - `(0, 10] -> 1, 2, 3, 4, 5, 6, 7, 8, 9, 10`
- Automatic Flattening.
    - Works with: RandomValue, TruffleShuffle, QuantumMonty, WeightedChoice & FlexCat.
    - Lazy Evaluation. All Random Value Generator Classes in Fortuna will recursively call or "flatten" callable objects returned from the data at call time.
    - Mixing callable objects with un-callable objects is fully supported, but it can look a bit messy.
    - Nested callable objects are fully supported. Because `lambda(lambda) -> lambda` fixes everything for arbitrary values of 'because', 'fixes' and 'everything'.
    - To disable automatic flattening, pass the optional keyword argument `flat=False` to the constructor.
    - **Under development:** It will soon be possible to auto-flatten callable objects that require multiple arguments. RandomValue already supports this feature.


## Random Value Generators

### Fortuna.RandomValue
`Fortuna.RandomValue(data: Collection, flat=True) -> Callable -> Value`
- @param data :: Collection of Values.
- @param flat :: Bool. Default: True. Option to automatically flatten callable values with lazy evaluation.
- @return :: Callable Object. `Callable(*args, zero_cool=random_index, range_to=0, **kwargs) -> Value`
    - @param zero_cool :: Optional ZeroCool Method, kwarg only. Default = random_index().
    - @param range_to :: Optional Integer in range [-N, N] where N is the length of the Collection. 
        - Default = 0, kwarg only. Parameter for ZeroCool Method.
        - Negative values of range_to indicate ranging from the back of the Collection. 
    - @param *args, **kwargs :: Optional arguments used to flatten the return Value.
    - @return Value or Value(*args, **kwargs) if it's Callable.

```python
from Fortuna import RandomValue, front_linear, back_linear

# Data Setup
random_apple = RandomValue(("Delicious", "Empire", "Granny Smith", "Honey Crisp", "Macintosh"))
random_fruit = RandomValue((
    lambda: f"Apple, {random_apple()}",
    "Banana",
    "Cherry",
    "Grapes",
    "Orange",
))

# Usage
print(random_fruit())
# prints a random fruit with the default flat uniform distribution

print(random_fruit(zero_cool=back_linear))
# prints a random fruit with a back_linear distribution

print(random_fruit(range_to=3))
# prints a random fruit of the first 3

print(random_fruit(zero_cool=front_linear, range_to=-3))
# prints a random fruit of the last 3 with a front_linear distribution of that range.
```


### TruffleShuffle
`Fortuna.TruffleShuffle(data: Collection, flat=True) -> Callable -> Value`
- Non-destructive, copies the data once.
- @param data :: Collection of Values. Set recommended but not required.
- @param flat :: Bool. Default: True. Option to automatically flatten callable values with lazy evaluation.
- @return :: Callable
    - @return :: Random value from the list with a **Wide Uniform Distribution**. The average width of the output distribution will naturally scale up with the size of the set.

**Wide Uniform Distribution**: *"Wide"* refers to the average distance between consecutive occurrences of the same value in the output sequence. The goal of this type of distribution is to keep the output sequence free of clumps or streaks of the same value, while maintaining randomness and uniform probability. This is not the same as a *flat uniform distribution*. The two distributions over time will be statistically similar for any given set, but the repetitiveness of the output sequence will be very different.

#### TruffleShuffle, Basic Use
```python
from Fortuna import TruffleShuffle

# Data Setup
list_of_values = { 1, 2, 3, 4, 5, 6 }
truffle_shuffle = TruffleShuffle(list_of_values)

# Usage
print(truffle_shuffle())  # this will print one of the numbers 1-6, 
# over time it will produce a wide distribution.
```

#### TruffleShuffle with Flattening
```python
from Fortuna import TruffleShuffle


auto_flat = TruffleShuffle([lambda: 1, lambda: 2, lambda: 3])
print(auto_flat())  # will print the value 1, 2 or 3.
# Note: the lambda will not be called until call time and stays dynamic for the life of the object.

un_flat = TruffleShuffle([lambda: 1, lambda: 2, lambda: 3], flat=False)
print(un_flat()())  # will print the value 1, 2 or 3, mind the double-double parenthesis

auto_un_flat = TruffleShuffle([lambda x: x, lambda x: x + 1, lambda x:  x + 2], flat=False)
# Note: flat=False is not required here because the lambdas can not be called without input x satisfied.
# It is still recommended to specify flat=False if non-flat output is intend.
print(auto_un_flat()(1))  # will print the value 1, 2 or 3, mind the double-double parenthesis
```

#### Mixing Static Objects with Callable Objects
```python
from Fortuna import TruffleShuffle


""" With automatic flattening active, lambda() -> int can be treated as an int. """
mixed_flat = TruffleShuffle([1, 2, lambda: 3])
print(mixed_flat())  # will print 1, 2 or 3

mixed_un_flat = TruffleShuffle([1, 2, lambda: 3], flat=False) # this pattern is not recommended.
print(mixed_flat())  # will print 1, 2 or "Function <lambda at some_address>"
# This pattern is not recommended because you wont know the nature of what you get back.
# This is almost always not what you want, and it can give rise to messy logic in other areas of your code.
```

#### Dynamic Strings
To successfully express a dynamic string, and keep it dynamic, at least one level of indirection is required. Without an indirection the f-string would collapse into a static string too soon.

```python
from Fortuna import TruffleShuffle, d


# d() is a simple dice function, d(n) -> [1, n] flat uniform distribution.
dynamic_string = TruffleShuffle((
    # while the probability of all A == all B == all C, individual probabilities of each possible string will differ based on the number of possible outputs of each category.
    lambda: f"A{d(2)}",  # -> A1 - A2, each are twice as likely as any particular B, and three times as likely as any C.
    lambda: f"B{d(4)}",  # -> B1 - B4, each are half as likely as any particular A, and 3/2 as likely as any C.
    lambda: f"C{d(6)}",  # -> C1 - C6, each are 1/3 as likely as any particular A and 2/3 as likely of any B.
))

print(dynamic_string())  # prints a random dynamic string, flattened at call time.

"""
>>> distribution_timer(dynamic_string)
Output Analysis: TruffleShuffle(sequence_data)()
Typical Timing: 875 ± 8 ns
Distribution of 10000 Samples:
 A1: 16.9%
 A2: 16.43%
 B1: 8.39%
 B2: 8.21%
 B3: 8.28%
 B4: 8.46%
 C1: 5.25%
 C2: 6.09%
 C3: 5.59%
 C4: 5.32%
 C5: 5.41%
 C6: 5.67%
"""
```

#### Nesting Dolls
```python
from Fortuna import TruffleShuffle

# Data Setup
nesting_dolls = TruffleShuffle({
    TruffleShuffle({"A", "B", "C", "D", "E"}),
    TruffleShuffle({"F", "G", "H", "I", "J"}),
    TruffleShuffle({"K", "L", "M", "N", "O"}),
    TruffleShuffle({"P", "Q", "R", "S", "T"}),
    ...
})

# Usage
print(nesting_dolls())  # prints one of the letters A-T, double wide distribution.
```

### QuantumMonty
`Fortuna.QuantumMonty(data: Collection, flat=True) -> Callable -> Value`
- @param data :: Collection of Values.
- @param flat :: Bool. Default: True. Option to automatically flatten callable values with lazy evaluation.
- @return :: Callable Object with Monty Methods for producing various distributions of the data.
    - @return :: Random value from the data. The instance will produce random values from the list using the selected distribution model or "monty". The default monty is the Quantum Monty Algorithm.

```python
from Fortuna import QuantumMonty

# Data Setup
list_of_values = [1, 2, 3, 4, 5, 6]
monty = QuantumMonty(list_of_values)

# Usage
print(monty())               # prints a random value from the list_of_values.
                             # uses the default Quantum Monty Algorithm.

print(monty.flat_uniform())  # prints a random value from the list_of_values.
                             # uses the "flat_uniform" monty.
                             # equivalent to random.choice(list_of_values).
```
The QuantumMonty class represents a diverse collection of strategies for producing random values from a sequence where the output distribution is based on the method you choose. Generally speaking, each value in the sequence will have a probability that is based on its position in the sequence. For example: the "front" monty produces random values where the beginning of the sequence is geometrically more common than the back. Given enough samples the "front" monty will always converge to a 45 degree slope down for any list of unique values.

There are three primary method families: linear, gaussian, and poisson. Each family has three base methods; 'front', 'middle', 'back', plus a 'quantum' method that incorporates all three base methods. The quantum algorithms for each family produce distributions by overlapping the probability waves of the other methods in their family. The Quantum Monty Algorithm incorporates all nine base methods.

```python
import Fortuna

# Data Setup
monty = Fortuna.QuantumMonty(
    ["Alpha", "Beta", "Delta", "Eta", "Gamma", "Kappa", "Zeta"]
)

# Usage
# Each of the following methods will return a random value from the sequence.
# Each method has its own unique distribution model.
""" Flat Base Case """
monty.flat_uniform()        # Flat Uniform Distribution
""" Geometric Positional """
monty.front_linear()        # Linear Descending, Triangle
monty.middle_linear()       # Linear Median Peak, Equilateral Triangle
monty.back_linear()         # Linear Ascending, Triangle
monty.quantum_linear()      # Linear Overlay, 3-way monty.
""" Gaussian Positional """
monty.front_gauss()         # Front Gamma
monty.middle_gauss()        # Scaled Gaussian
monty.back_gauss()          # Reversed Gamma
monty.quantum_gauss()       # Gaussian Overlay, 3-way monty.
""" Poisson Positional """
monty.front_poisson()       # 1/4 Mean Poisson
monty.middle_poisson()      # 1/2 Mean Poisson
monty.back_poisson()        # 3/4 Mean Poisson
monty.quantum_poisson()     # Poisson Overlay, 3-way monty.
""" Quantum Monty Algorithm """
monty()                     # Quantum Monty Algorithm, 9-way monty.
monty.quantum_monty()       #  same as above
```

### Weighted Choice: Base Class
Weighted Choice offers two strategies for selecting random values from a sequence where programmable rarity is desired. Both produce a custom distribution of values based on the weights of the values.

The choice to use one strategy over the other is purely about which one suits you or your data best. Relative weights are easier to understand at a glance. However, many RPG Treasure Tables map rather nicely to a cumulative weighted strategy.

#### Cumulative Weighted Choice
`Fortuna.CumulativeWeightedChoice(weighted_table: Table, flat=True) -> Callable -> Value`
- @param weighted_table :: Table of weighted pairs.
- @param flat :: Bool. Default: True. Option to automatically flatten callable values with lazy evaluation.
- @return :: Callable Instance
    - @return :: Random value from the weighted_table, distribution based on the weights of the values.

_Note: Logic dictates Cumulative Weights must be unique!_

```python
from Fortuna import CumulativeWeightedChoice

# Data Setup
cum_weighted_choice = CumulativeWeightedChoice((
    (7, "Apple"),
    (11, "Banana"),
    (13, "Cherry"),
    (23, "Grape"),
    (26, "Lime"),
    (30, "Orange"),  # same as relative weight 4 because 30 - 26 = 4
))
# Usage
print(cum_weighted_choice())  # prints a weighted random value
```

#### Relative Weighted Choice
`Fortuna.RelativeWeightedChoice(weighted_table: Table) -> Callable -> Value`
- @param weighted_table :: Table of weighted pairs.
- @param flat :: Bool. Default: True. Option to automatically flatten callable values with lazy evaluation.
- @return :: Callable Instance
    - @return :: Random value from the weighted_table, distribution based on the weights of the values.

```python
from Fortuna import RelativeWeightedChoice

# Data
population = ["Apple", "Banana", "Cherry", "Grape", "Lime", "Orange"]
rel_weights = [7, 4, 2, 10, 3, 4]

# Setup
rel_weighted_choice = RelativeWeightedChoice(zip(rel_weights, population))

# Usage
print(rel_weighted_choice())  # prints a weighted random value
```

### FlexCat
`Fortuna.FlexCat(dict_of_lists: Matrix, key_bias="front_linear", val_bias="truffle_shuffle", flat=True) -> Callable -> Value`
- @param dict_of_lists :: Keyed Matrix of Value Sequences.
- @parm key_bias :: String indicating the name of the algorithm to use for random key selection.
- @parm val_bias :: String indicating the name of the algorithm to use for random value selection.
- @param flat :: Bool. Default: True. Option to automatically flatten callable values with lazy evaluation.
- @return :: Callable Instance
    - @param cat_key :: Optional String. Default is None. Key selection by name. If specified, this will override the key_bias for a single call.
    - @return :: Value. Returns a random value generated with val_bias from a random sequence generated with key_bias.

FlexCat is like a two dimensional QuantumMonty, or a QuantumMonty of QuantumMontys.

The constructor takes two optional keyword arguments to specify the algorithms to be used to make random selections. The algorithm specified for selecting a key need not be the same as the one for selecting values. An optional key may be provided at call time to bypass the random key selection. Keys passed in this way must exactly match a key in the Matrix.

By default, FlexCat will use key_bias="front_linear" and val_bias="truffle_shuffle", this will make the top of the data structure geometrically more common than the bottom and it will truffle shuffle the sequence values. This config is known as TopCat, it produces a descending-step, micro-shuffled distribution sequence. Many other combinations are available.

Algorithmic Options: _See QuantumMonty & TruffleShuffle for more details._
- "front_linear", Linear Descending
- "middle_linear", Linear Median Peak
- "back_linear", Linear Ascending
- "quantum_linear", Linear 3-way monty
- "front_gauss", Gamma Descending
- "middle_gauss", Scaled Gaussian
- "back_gauss", Gamma Ascending
- "quantum_gauss", Gaussian 3-way monty
- "front_poisson", Front 1/3 Mean Poisson
- "middle_poisson", Middle Mean Poisson
- "back_poisson", Back 1/3 Mean Poisson
- "quantum_poisson", Poisson 3-way monty
- "quantum_monty", Quantum Monty Algorithm, 9-way monty
- "flat_uniform", uniform flat distribution
- "truffle_shuffle", TruffleShuffle, wide uniform distribution

```python
from Fortuna import FlexCat, d
#
#
#                           |- Collection Generator, does not require lambda.
#                           |  Cat_A is equivalent to Cat_B
# Data                      |
matrix_data = {#            $                         |- Dynamic Value Expression
    "Cat_A": (f"A{i}" for i in range(1, 6)),  #       |  Lazy evaluation, 1 of 4
    "Cat_B": ("B1", "B2", "B3", "B4", "B5"),  #       $
    "Cat_C": ("C1", "C2", "C3", f"C4.{d(2)}", lambda: f"C5.{d(4)}")}
#    $       $       $              $                        $
#    |       |       |- Value       |                        |- Fair die method
#    |       |                      |
#    |       |- Collection          |- Static Value Expression, evaluated only once.
#    |                              |  Eager evaluation, 1 of 2 permanently
#    |- Collection Key, "cat_key"
#                                   |- Collection Selection  |- Value Selection
# Setup                             $  Y-axis                $  X-axis
flex_cat = FlexCat(matrix_data, key_bias="front_linear", val_bias="flat_uniform")
#    $       $       $
#    |       |       |- Matrix, Dictionary of Collections
#    |       |
#    |       |- Constructor
#    |       
#    |- Callable
#
# Usage
flex_cat()  # returns a Value from the Matrix.
flex_cat(cat_key="Cat_B")  # returns a Value specifically from the "Cat_B" Collection.
```

### Random Integer Generators
`Fortuna.random_below(number: int) -> int`
- @param number :: Any Integer
- @return :: Returns a random integer in the range...
    - `random_below(number) -> [0, number)` for positive values.
    - `random_below(number) -> (number, 0]` for negative values.
    - `random_below(0) -> 0` Always returns zero when input is zero
- Flat uniform distribution.


`Fortuna.random_int(left_limit: int, right_limit: int) -> int`
- @param left_limit :: Any Integer
- @param right_limit :: Any Integer
- @return :: Returns a random integer in the range [left_limit, right_limit]
    - `random_int(1, 10) -> [1, 10]`
    - `random_int(10, 1) -> [1, 10]` same as above.
    - `random_int(A, B)` Always returns A when A == B
- Flat uniform distribution.


`Fortuna.random_range(start: int, stop: int = 0, step: int = 1) -> int`
- @param start :: Required starting point.
    - `random_range(0) -> [0]`
    - `random_range(10) -> [0, 10)` from 0 to 9. Same as `Fortuna.random_index(N)`
    - `random_range(-10) -> [-10, 0)` from -10 to -1. Same as `Fortuna.random_index(-N)`
- @param stop :: Zero by default. Optional range bound. With at least two arguments, the order of the first two does not matter.
    - `random_range(0, 0) -> [0]`
    - `random_range(0, 10) -> [0, 10)` from 0 to 9.
    - `random_range(10, 0) -> [0, 10)` same as above.
- @param step :: One by default. Optional step size.
    - `random_range(0, 0, 0) -> [0]`
    - `random_range(0, 10, 2) -> [0, 10) by 2` even numbers from 0 to 8.
    - The sign of the step parameter controls the phase of the output. Negative stepping will flip the inclusively.
    - `random_range(0, 10, -1) -> (0, 10]` starts at 10 and ranges down to 1.
    - `random_range(10, 0, -1) -> (0, 10]` same as above.
    - `random_range(10, 10, 0) -> [10]` a step size or range size of zero always returns the first parameter.
- @return :: Returns a random integer in the range [A, B) by increments of C.
- Flat uniform distribution.


`Fortuna.d(sides: int) -> int`
- Represents a single roll of a given size die.
- @param sides :: Any Integer. Represents the size or number of sides, most commonly six.
- @return :: Returns a random integer in the range [1, sides].
- Flat uniform distribution.


`Fortuna.dice(rolls: int, sides: int) -> int`
- Represents the sum of multiple rolls of the same size die.
- @param rolls :: Any Integer. Represents the number of times to roll the die.
- @param sides :: Any Integer. Represents the die size or number of sides, most commonly six.
- @return :: Returns a random integer in range [X, Y] where X = rolls and Y = rolls * sides.
- Geometric distribution based on the number and size of the dice rolled.
- Complexity scales primarily with the number of rolls, not the size of the dice.


`Fortuna.plus_or_minus(number: int) -> int`
- @param number :: Any Integer.
- @return :: Returns a random integer in range [-number, number].
- Flat uniform distribution.


`Fortuna.plus_or_minus_linear(number: int) -> int`
- @param number :: Any Integer.
- @return :: Returns a random integer in range [-number, number].
- Linear geometric, 45 degree triangle distribution centered on zero.


`Fortuna.plus_or_minus_gauss(number: int) -> int`
- @param number :: Any Integer.
- @return :: Returns a random integer in range [-number, number].
- Stretched gaussian distribution centered on zero.


### Random Index, ZeroCool Specification
ZeroCool Methods are used to generate random Sequence indices.

ZeroCool methods must have the following properties:
- Any distribution model is acceptable.
- The method or function must take exactly one Integer parameter N.
- The method returns a random int in range `[0, N)` for positive values of N.
- The method returns a random int in range `[N, 0)` for negative values of N.
    - This symmetry matches how python can index a list from the back for negative index values or from the front for positive index values.

ZeroCool functions often have an interesting limit as size goes to zero. ZeroCool does not place requirements the output of this input limit. At higher levels of abstraction, inside classes that employ ZeroCool methods- zero is always a sentinel to indicate the full range. In that case the length of the list is sent to the ZeroCool method, not zero. However for those who enjoy thinking a little deeper, consider the following:

If given the fact that an empty range is never an option, we could design a better solution than failure for input zero. Calculus might suggest that both infinity and negative infinity are equally viable output for an input limit of zero, but both are inappropriate for indexing a list. However, if we map infinity to the back of the list and minus infinity to the front of the list, then the following might hold: `random_index(0) -> [-1, 0]`. This "Top or Bottom" solution is not required for a method to be ZeroCool compatible, it is just an interesting option. Other valid possibilities include: always return None or 0 or -1 or throw an exception or spawn nasal demons, however none of these seem terribly helpful or useful. At least the Top/Bottom solution accurately reflects the "off by one" symmetry of the domain mapping that defines ZeroCool methods in general.


```python
from Fortuna import random_index


some_list = [i for i in range(100)]

print(some_list[random_index(10)])  # prints one of the first 10 items of some_list, [0, 9]
print(some_list[random_index(-10)])  # prints one of the last 10 items of some_list, [90, 99]
```
### ZeroCool Methods
- `Fortuna.random_index(size: int) -> int` Flat uniform distribution
- `Fortuna.front_gauss(size: int) -> int` Gamma Distribution: Front Peak
- `Fortuna.middle_gauss(size: int) -> int` Stretched Gaussian Distribution: Median Peak
- `Fortuna.back_gauss(size: int) -> int` Gamma Distribution: Back Peak
- `Fortuna.quantum_gauss(size: int) -> int` Quantum Gaussian: Three-way Monty
- `Fortuna.front_poisson(size: int) -> int` Poisson Distribution: Front 1/3 Peak
- `Fortuna.middle_poisson(size: int) -> int` Poisson Distribution: Middle Peak
- `Fortuna.back_poisson(size: int) -> int` Poisson Distribution: Back 1/3 Peak
- `Fortuna.quantum_poisson(size: int) -> int` Quantum Poisson: Three-way Monty
- `Fortuna.front_geometric(size: int) -> int` Linear Geometric: 45 Degree Front Peak
- `Fortuna.middle_geometric(size: int) -> int` Linear Geometric: 45 Degree Middle Peak
- `Fortuna.back_geometric(size: int) -> int` Linear Geometric: 45 Degree Back Peak
- `Fortuna.quantum_geometric(size: int) -> int` Quantum Geometric: Three-way Monty
- `Fortuna.quantum_monty(size: int) -> int` Quantum Monty: Twelve-way Monty

```python
from Fortuna import front_gauss, middle_gauss, back_gauss, quantum_gauss


some_list = [i for i in range(100)]

# Each of the following prints one of the first 10 items of some_list with the appropriate distribution
print(some_list[front_gauss(10)])
print(some_list[middle_gauss(10)])
print(some_list[back_gauss(10)])
print(some_list[quantum_gauss(10)])

# Each of the following prints one of the last 10 items of some_list with the appropriate distribution
print(some_list[front_gauss(-10)])  
print(some_list[middle_gauss(-10)])  
print(some_list[back_gauss(-10)])  
print(some_list[quantum_gauss(-10)])
```

### Random Float Generators
`Fortuna.canonical() -> float`
- @return :: random float in range [0.0, 1.0), flat uniform.


`Fortuna.random_float(a: Float, b: Float) -> Float`
- @param a :: Float
- @param b :: Float
- @return :: random Float in range [a, b), flat uniform distribution.


### Random Truth Generator
`Fortuna.percent_true(truth_factor: Float = 50.0) -> bool`
- @param truth_factor :: The probability of True as a percentage. Default is 50 percent.
- @return :: Produces True or False based on the truth_factor.
    - Always returns False if num is 0 or less
    - Always returns True if num is 100 or more.


### Shuffle Algorithms
`Fortuna.shuffle(array: list) -> None`
- Knuth B shuffle algorithm. Destructive, in-place shuffle.
- @param array :: Must be a mutable list.

`Fortuna.knuth(array: list) -> None`
- Knuth A shuffle algorithm. Destructive, in-place shuffle.
- @param array :: Must be a mutable list.

`Fortuna.fisher_yates(array: list) -> None`
- Fisher-Yates shuffle algorithm. Destructive, in-place shuffle.
- @param array :: Must be a mutable list.


### Test Suite
`Fortuna.distribution_timer(func: staticmethod, *args, num_cycles=100000, **kwargs)`

`Fortuna.quick_test(num_cycles=10000)`


## Fortuna Development Log
##### Fortuna 3.6.5
- Documentation Update
- RandomValue: New flatten-with-arguments functionality.

##### Fortuna 3.6.4
- RandomValue added for testing

##### Fortuna 3.6.3
- Developer Update

##### Fortuna 3.6.2
- Installer Script Update

##### Fortuna 3.6.1
- Documentation Update

##### Fortuna 3.6.0
- Storm Update
- Test Update
- Bug fix for random_range(), negative stepping is now working as intended. This bug was introduced in 3.5.0.
- Removed Features
    - lazy_cat(): use QuantumMonty class instead.
    - flex_cat(): use FlexCat class instead.
    - truffle_shuffle(): use TruffleShuffle class instead.

##### Fortuna 3.5.3 - internal
- Features added for testing & development
    - ActiveChoice class
    - random_rotate() function

##### Fortuna 3.5.2
- Documentation Updates

##### Fortuna 3.5.1
- Test Update

##### Fortuna 3.5.0
- Storm Update
- Minor Bug Fix: Truffle Shuffle
- Deprecated Features
    - lazy_cat(): use QuantumMonty class instead.
    - flex_cat(): use FlexCat class instead.
    - truffle_shuffle(): use TruffleShuffle class instead.

##### Fortuna 3.4.9
- Test Update

##### Fortuna 3.4.8
- Storm Update

##### Fortuna 3.4.7
- Bug fix for analytic_continuation.

##### Fortuna 3.4.6
- Docs Update

##### Fortuna 3.4.5
- Docs Update
- Range Tests Added, see extras folder.

##### Fortuna 3.4.4
- ZeroCool Algorithm Bug Fixes
- Typos Fixed

##### Fortuna 3.4.3
- Docs Update

##### Fortuna 3.4.2
- Typos Fixed

##### Fortuna 3.4.1
- Major Bug Fix: random_index()

##### Fortuna 3.4.0 - internal
- ZeroCool Poisson Algorithm Family Updated

##### Fortuna 3.3.8 - internal
- Docs Update

##### Fortuna 3.3.7
- Fixed Performance Bug: ZeroCool Linear Algorithm Family

##### Fortuna 3.3.6
- Docs Update

##### Fortuna 3.3.5
- ABI Updates
- Bug Fixes

##### Fortuna 3.3.4
- Examples Update

##### Fortuna 3.3.3
- Test Suite Update

##### Fortuna 3.3.2 - internal
- Documentation Update

##### Fortuna 3.3.1 - internal
- Minor Bug Fix

##### Fortuna 3.3.0 - internal
- Added `plus_or_minus_gauss(N: int) -> int` random int in range [-N, N] Stretched Gaussian Distribution

##### Fortuna 3.2.3
- Small Typos Fixed

##### Fortuna 3.2.2
- Documentation update.

##### Fortuna 3.2.1
- Small Typo Fixed

##### Fortuna 3.2.0
- API updates:
    - QunatumMonty.uniform -> QunatumMonty.flat_uniform
    - QunatumMonty.front -> QunatumMonty.front_linear
    - QunatumMonty.middle -> QunatumMonty.middle_linear
    - QunatumMonty.back -> QunatumMonty.back_linear
    - QunatumMonty.quantum -> QunatumMonty.quantum_linear
    - randindex -> random_index
    - randbelow -> random_below
    - randrange -> random_range
    - randint   -> random_int

##### Fortuna 3.1.0
- `discrete()` has been removed, see Weighted Choice.
- `lazy_cat()` added.
- All ZeroCool methods have been raised to top level API, for use with lazy_cat()

##### Fortuna 3.0.1
- minor typos.

##### Fortuna 3.0.0
- Storm 2 Rebuild.

##### Fortuna 2.1.1
- Small bug fixes.
- Test updates.

##### Fortuna 2.1.0, Major Feature Update
- Fortuna now includes the best of RNG and Pyewacket.

##### Fortuna 2.0.3
- Bug fix.

##### Fortuna 2.0.2
- Clarified some documentation.

##### Fortuna 2.0.1
- Fixed some typos.

##### Fortuna 2.0.0b1-10
- Total rebuild. New RNG Storm Engine.

##### Fortuna 1.26.7.1
- README updated.

##### Fortuna 1.26.7
- Small bug fix.

##### Fortuna 1.26.6
- Updated README to reflect recent changes to the test script.

##### Fortuna 1.26.5
- Fixed small bug in test script.

##### Fortuna 1.26.4
- Updated documentation for clarity.
- Fixed a minor typo in the test script.

##### Fortuna 1.26.3
- Clean build.

##### Fortuna 1.26.2
- Fixed some minor typos.

##### Fortuna 1.26.1
- Release.

##### Fortuna 1.26.0 beta 2
- Moved README and LICENSE files into fortuna_extras folder.

##### Fortuna 1.26.0 beta 1
- Dynamic version scheme implemented.
- The Fortuna Extension now requires the fortuna_extras package, previously it was optional.

##### Fortuna 1.25.4
- Fixed some minor typos in the test script.

##### Fortuna 1.25.3
- Since version 1.24 Fortuna requires Python 3.7 or higher. This patch corrects an issue where the setup script incorrectly reported requiring Python 3.6 or higher.

##### Fortuna 1.25.2
- Updated test suite.
- Major performance update for TruffleShuffle.
- Minor performance update for QuantumMonty & FlexCat: cycle monty.

##### Fortuna 1.25.1
- Important bug fix for TruffleShuffle, QuantumMonty and FlexCat.

##### Fortuna 1.25
- Full 64bit support.
- The Distribution & Performance Tests have been redesigned.
- Bloat Control: Two experimental features have been removed.
    - RandomWalk
    - CatWalk
- Bloat Control: Several utility functions have been removed from the top level API. These function remain in the Fortuna namespace for now, but may change in the future without warning.
    - stretch_bell, internal only.
    - min_max, not used anymore.
    - analytic_continuation, internal only.
    - flatten, internal only.

##### Fortuna 1.24.3
- Low level refactoring, non-breaking patch.

##### Fortuna 1.24.2
- Setup config updated to improve installation.

##### Fortuna 1.24.1
- Low level patch to avoid potential ADL issue. All low level function calls are now qualified.

##### Fortuna 1.24
- Documentation updated for even more clarity.
- Bloat Control: Two naïve utility functions that are no longer used in the module have been removed.
    - n_samples -> use a list comprehension instead. `[f(x) for _ in range(n)]`
    - bind -> use a lambda instead. `lambda: f(x)`

##### Fortuna 1.23.7
- Documentation updated for clarity.
- Minor bug fixes.
- TruffleShuffle has been redesigned slightly, it now uses a random rotate instead of swap.
- Custom `__repr__` methods have been added to each class.

##### Fortuna 1.23.6
- New method for QuantumMonty: quantum_not_monty - produces the upside down quantum_monty.
- New bias option for FlexCat: not_monty.

##### Fortuna 1.23.5.1
- Fixed some small typos.

##### Fortuna 1.23.5
- Documentation updated for clarity.
- All sequence wrappers can now accept generators as input.
- Six new functions added:
    - random_float() -> float in range [0.0..1.0) exclusive, uniform flat distribution.
    - percent_true_float(num: float) -> bool, Like percent_true but with floating point precision.
    - plus_or_minus_linear_down(num: int) -> int in range [-num..num], upside down pyramid.
    - plus_or_minus_curve_down(num: int) -> int in range [-num..num], upside down bell curve.
    - mostly_not_middle(num: int) -> int in range [0..num], upside down pyramid.
    - mostly_not_center(num: int) -> int in range [0..num], upside down bell curve.
- Two new methods for QuantumMonty:
    - mostly_not_middle
    - mostly_not_center
- Two new bias options for FlexCat, either can be used to define x and/or y axis bias:
    - not_middle
    - not_center

##### Fortuna 1.23.4.2
- Fixed some minor typos in the README.md file.

##### Fortuna 1.23.4.1
- Fixed some minor typos in the test suite.

##### Fortuna 1.23.4
- Fortuna is now Production/Stable!
- Fortuna and Fortuna Pure now use the same test suite.

##### Fortuna 0.23.4, first release candidate.
- RandomCycle, BlockCycle and TruffleShuffle have been refactored and combined into one class: TruffleShuffle.
- QuantumMonty and FlexCat will now use the new TruffleShuffle for cycling.
- Minor refactoring across the module.

##### Fortuna 0.23.3, internal
- Function shuffle(arr: list) added.

##### Fortuna 0.23.2, internal
- Simplified the plus_or_minus_curve(num: int) function, output will now always be bounded to the range [-num..num].
- Function stretched_bell(num: int) added, this matches the previous behavior of an unbounded plus_or_minus_curve.

##### Fortuna 0.23.1, internal
- Small bug fixes and general clean up.

##### Fortuna 0.23.0
- The number of test cycles in the test suite has been reduced to 10,000 (down from 100,000). The performance of the pure python implementation and the c-extension are now directly comparable.
- Minor tweaks made to the examples in `.../fortuna_extras/fortuna_examples.py`

##### Fortuna 0.22.2, experimental features
- BlockCycle class added.
- RandomWalk class added.
- CatWalk class added.

##### Fortuna 0.22.1
- Fortuna classes no longer return lists of values, this behavior has been extracted to a free function called n_samples.

##### Fortuna 0.22.0, experimental features
- Function bind added.
- Function n_samples added.

##### Fortuna 0.21.3
- Flatten will no longer raise an error if passed a callable item that it can't call. It correctly returns such items in an uncalled state without error.
- Simplified `.../fortuna_extras/fortuna_examples.py` - removed unnecessary class structure.

##### Fortuna 0.21.2
- Fix some minor bugs.

##### Fortuna 0.21.1
- Fixed a bug in `.../fortuna_extras/fortuna_examples.py`

##### Fortuna 0.21.0
- Function flatten added.
- Flatten: The Fortuna classes will recursively unpack callable objects in the data set.

##### Fortuna 0.20.10
- Documentation updated.

##### Fortuna 0.20.9
- Minor bug fixes.

##### Fortuna 0.20.8, internal
- Testing cycle for potential new features.

##### Fortuna 0.20.7
- Documentation updated for clarity.

##### Fortuna 0.20.6
- Tests updated based on recent changes.

##### Fortuna 0.20.5, internal
- Documentation updated based on recent changes.

##### Fortuna 0.20.4, internal
- WeightedChoice (both types) can optionally return a list of samples rather than just one value, control the length of the list via the n_samples argument.

##### Fortuna 0.20.3, internal
- RandomCycle can optionally return a list of samples rather than just one value,
control the length of the list via the n_samples argument.

##### Fortuna 0.20.2, internal
- QuantumMonty can optionally return a list of samples rather than just one value,
control the length of the list via the n_samples argument.

##### Fortuna 0.20.1, internal
- FlexCat can optionally return a list of samples rather than just one value,
control the length of the list via the n_samples argument.

##### Fortuna 0.20.0, internal
- FlexCat now accepts a standard dict as input. The ordered(ness) of dict is now part of the standard in Python 3.7.1. Previously FlexCat required an OrderedDict, now it accepts either and treats them the same.

##### Fortuna 0.19.7
- Fixed bug in `.../fortuna_extras/fortuna_examples.py`.

##### Fortuna 0.19.6
- Updated documentation formatting.
- Small performance tweak for QuantumMonty and FlexCat.

##### Fortuna 0.19.5
- Minor documentation update.

##### Fortuna 0.19.4
- Minor update to all classes for better debugging.

##### Fortuna 0.19.3
- Updated plus_or_minus_curve to allow unbounded output.

##### Fortuna 0.19.2
- Internal development cycle.
- Minor update to FlexCat for better debugging.

##### Fortuna 0.19.1
- Internal development cycle.

##### Fortuna 0.19.0
- Updated documentation for clarity.
- MultiCat has been removed, it is replaced by FlexCat.
- Mostly has been removed, it is replaced by QuantumMonty.

##### Fortuna 0.18.7
- Fixed some more README typos.

##### Fortuna 0.18.6
- Fixed some README typos.

##### Fortuna 0.18.5
- Updated documentation.
- Fixed another minor test bug.

##### Fortuna 0.18.4
- Updated documentation to reflect recent changes.
- Fixed some small test bugs.
- Reduced default number of test cycles to 10,000 - down from 100,000.

##### Fortuna 0.18.3
- Fixed some minor README typos.

##### Fortuna 0.18.2
- Fixed a bug with Fortuna Pure.

##### Fortuna 0.18.1
- Fixed some minor typos.
- Added tests for `.../fortuna_extras/fortuna_pure.py`

##### Fortuna 0.18.0
- Introduced new test format, now includes average call time in nanoseconds.
- Reduced default number of test cycles to 100,000 - down from 1,000,000.
- Added pure Python implementation of Fortuna: `.../fortuna_extras/fortuna_pure.py`
- Promoted several low level functions to top level.
    - `zero_flat(num: int) -> int`
    - `zero_cool(num: int) -> int`
    - `zero_extreme(num: int) -> int`
    - `max_cool(num: int) -> int`
    - `max_extreme(num: int) -> int`
    - `analytic_continuation(func: staticmethod, num: int) -> int`
    - `min_max(num: int, lo: int, hi: int) -> int`

##### Fortuna 0.17.3
- Internal development cycle.

##### Fortuna 0.17.2
- User Requested: dice() and d() functions now support negative numbers as input.

##### Fortuna 0.17.1
- Fixed some minor typos.

##### Fortuna 0.17.0
- Added QuantumMonty to replace Mostly, same default behavior with more options.
- Mostly is depreciated and may be removed in a future release.
- Added FlexCat to replace MultiCat, same default behavior with more options.
- MultiCat is depreciated and may be removed in a future release.
- Expanded the Treasure Table example in `.../fortuna_extras/fortuna_examples.py`

##### Fortuna 0.16.2
- Minor refactoring for WeightedChoice.

##### Fortuna 0.16.1
- Redesigned fortuna_examples.py to feature a dynamic random magic item generator.
- Raised cumulative_weighted_choice function to top level.
- Added test for cumulative_weighted_choice as free function.
- Updated MultiCat documentation for clarity.

##### Fortuna 0.16.0
- Pushed distribution_timer to the .pyx layer.
- Changed default number of iterations of tests to 1 million, up form 1 hundred thousand.
- Reordered tests to better match documentation.
- Added Base Case Fortuna.fast_rand_below.
- Added Base Case Fortuna.fast_d.
- Added Base Case Fortuna.fast_dice.

##### Fortuna 0.15.10
- Internal Development Cycle

##### Fortuna 0.15.9
- Added Base Cases for random_value()
- Added Base Case for randint()

##### Fortuna 0.15.8
- Clarified MultiCat Test

##### Fortuna 0.15.7
- Fixed minor typos.

##### Fortuna 0.15.6
- Fixed minor typos.
- Simplified MultiCat example.

##### Fortuna 0.15.5
- Added MultiCat test.
- Fixed some minor typos in docs.

##### Fortuna 0.15.4
- Performance optimization for both WeightedChoice() variants.
- Cython update provides small performance enhancement across the board.
- Compilation now leverages Python3 all the way down.
- MultiCat pushed to the .pyx layer for better performance.

##### Fortuna 0.15.3
- Reworked the MultiCat example to include several randomizing strategies working in concert.
- Added Multi Dice 10d10 performance tests.
- Updated sudo code in documentation to be more pythonic.

##### Fortuna 0.15.2
- Fixed: Linux installation failure.
- Added: complete source files to the distribution (.cpp .hpp .pyx).

##### Fortuna 0.15.1
- Updated & simplified distribution_timer in `fortuna_tests.py`
- Readme updated, fixed some typos.
- Known issue preventing successful installation on some linux platforms.

##### Fortuna 0.15.0
- Performance tweaks.
- Readme updated, added some details.

##### Fortuna 0.14.1
- Readme updated, fixed some typos.

##### Fortuna 0.14.0
- Fixed a bug where the analytic continuation algorithm caused a rare issue during compilation on some platforms.

##### Fortuna 0.13.3
- Fixed Test Bug: percent sign was missing in output distributions.
- Readme updated: added update history, fixed some typos.

##### Fortuna 0.13.2
- Readme updated for even more clarity.

##### Fortuna 0.13.1
- Readme updated for clarity.

##### Fortuna 0.13.0
- Minor Bug Fixes.
- Readme updated for aesthetics.
- Added Tests: `.../fortuna_extras/fortuna_tests.py`

##### Fortuna 0.12.0
- Internal test for future update.

##### Fortuna 0.11.0
- Initial Release: Public Beta

##### Fortuna 0.10.0
- Module name changed from Dice to Fortuna

##### Dice 0.1.x - 0.9.x
- Experimental Phase


## Fortuna Distribution and Performance Test Suite
```
Fortuna Test Suite: RNG Storm Engine

Numeric Input Limits:

Internal 64 bit Integer Type <long long>
Min Integer: -9223372036854775807
Max Integer:  9223372036854775807

Internal 64 bit Float Type <double>
Min Float:   -1.7976931348623157e+308
Max Float:    1.7976931348623157e+308
Neg Epsilon: -5e-324
Pos Epsilon:  5e-324

Output Range Tests: Success!

-------------------------------------------------------------------------

Fortuna Quick Test

Random Sequence Values:

some_list = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

Base Case
Output Analysis: Random.choice(some_list)
Typical Timing: 750 ± 21 ns
Statistics of 256 Samples:
 Minimum: 0
 Median: 5
 Maximum: 9
 Mean: 4.81640625
 Std Deviation: 2.7881901458142497
Distribution of 100000 Samples:
 0: 10.021%
 1: 10.038%
 2: 9.957%
 3: 10.003%
 4: 10.068%
 5: 9.897%
 6: 9.962%
 7: 9.958%
 8: 10.033%
 9: 10.063%

Output Analysis: random_value(some_list)
Typical Timing: 63 ± 1 ns
Statistics of 256 Samples:
 Minimum: 0
 Median: 4
 Maximum: 9
 Mean: 4.19921875
 Std Deviation: 2.819289191288763
Distribution of 100000 Samples:
 0: 10.045%
 1: 10.081%
 2: 10.153%
 3: 10.097%
 4: 9.906%
 5: 9.862%
 6: 10.037%
 7: 9.962%
 8: 9.974%
 9: 9.883%

Output Analysis: TruffleShuffle(collection)()
Typical Timing: 500 ± 15 ns
Statistics of 256 Samples:
 Minimum: 0
 Median: 5
 Maximum: 9
 Mean: 4.546875
 Std Deviation: 2.8816100483216385
Distribution of 100000 Samples:
 0: 10.017%
 1: 9.979%
 2: 10.003%
 3: 10.003%
 4: 9.89%
 5: 10.015%
 6: 10.055%
 7: 10.047%
 8: 9.977%
 9: 10.014%

Output Analysis: QuantumMonty(collection)()
Typical Timing: 313 ± 12 ns
Statistics of 256 Samples:
 Minimum: 0
 Median: 5
 Maximum: 9
 Mean: 4.70703125
 Std Deviation: 2.6825707188748456
Distribution of 100000 Samples:
 0: 7.251%
 1: 9.114%
 2: 10.545%
 3: 11.374%
 4: 11.546%
 5: 11.831%
 6: 11.3%
 7: 10.532%
 8: 9.181%
 9: 7.326%

Output Analysis: RandomValue(collection)()
Typical Timing: 375 ± 1 ns
Statistics of 256 Samples:
 Minimum: 0
 Median: 5
 Maximum: 9
 Mean: 4.60546875
 Std Deviation: 2.9406897575228843
Distribution of 100000 Samples:
 0: 9.842%
 1: 10.171%
 2: 10.026%
 3: 9.908%
 4: 9.95%
 5: 10.114%
 6: 9.961%
 7: 10.055%
 8: 9.889%
 9: 10.084%


Weighted Tables:

population = ('A', 'B', 'C', 'D')
cum_weights = (1, 3, 6, 10)
rel_weights = (1, 2, 3, 4)
cum_weighted_table = zip(cum_weights, population)
rel_weighted_table = zip(rel_weights, population)

Cumulative Base Case
Output Analysis: Random.choices(population, cum_weights=cum_weights)
Typical Timing: 1688 ± 17 ns
Distribution of 100000 Samples:
 A: 9.971%
 B: 19.898%
 C: 30.106%
 D: 40.025%

Output Analysis: CumulativeWeightedChoice(weighted_table)()
Typical Timing: 250 ± 11 ns
Distribution of 100000 Samples:
 A: 9.926%
 B: 19.943%
 C: 30.048%
 D: 40.083%

Output Analysis: cumulative_weighted_choice(tuple(zip(cum_weights, population)))
Typical Timing: 125 ± 15 ns
Distribution of 100000 Samples:
 A: 9.972%
 B: 20.052%
 C: 29.763%
 D: 40.213%

Relative Base Case
Output Analysis: Random.choices(population, weights=rel_weights)
Typical Timing: 2063 ± 21 ns
Distribution of 100000 Samples:
 A: 9.909%
 B: 20.152%
 C: 30.01%
 D: 39.929%

Output Analysis: RelativeWeightedChoice(weighted_table)()
Typical Timing: 250 ± 15 ns
Distribution of 100000 Samples:
 A: 9.881%
 B: 20.166%
 C: 29.873%
 D: 40.08%


Random Matrix Values:

some_matrix = {'A': (1, 2, 3, 4), 'B': (10, 20, 30, 40), 'C': (100, 200, 300, 400)}

Output Analysis: FlexCat(matrix_data, key_bias='flat_uniform', val_bias='flat_uniform')()
Typical Timing: 407 ± 14 ns
Statistics of 256 Samples:
 Minimum: 1
 Median: 30
 Maximum: 400
 Mean: 97.7421875
 Std Deviation: 131.40723984554097
Distribution of 100000 Samples:
 1: 8.304%
 2: 8.461%
 3: 8.252%
 4: 8.212%
 10: 8.385%
 20: 8.217%
 30: 8.367%
 40: 8.342%
 100: 8.369%
 200: 8.505%
 300: 8.311%
 400: 8.275%


Random Integers:

Base Case
Output Analysis: Random.randrange(10)
Typical Timing: 844 ± 21 ns
Statistics of 256 Samples:
 Minimum: 0
 Median: 4
 Maximum: 9
 Mean: 4.5703125
 Std Deviation: 2.8824923685765347
Distribution of 100000 Samples:
 0: 10.009%
 1: 10.044%
 2: 9.936%
 3: 9.914%
 4: 10.014%
 5: 9.983%
 6: 10.017%
 7: 10.019%
 8: 10.038%
 9: 10.026%

Output Analysis: random_below(10)
Typical Timing: 63 ± 6 ns
Statistics of 256 Samples:
 Minimum: 0
 Median: 5
 Maximum: 9
 Mean: 4.57421875
 Std Deviation: 2.8796717158690974
Distribution of 100000 Samples:
 0: 10.063%
 1: 9.942%
 2: 9.966%
 3: 10.014%
 4: 10.078%
 5: 9.97%
 6: 9.923%
 7: 9.948%
 8: 10.056%
 9: 10.04%

Output Analysis: random_index(10)
Typical Timing: 63 ± 6 ns
Statistics of 256 Samples:
 Minimum: 0
 Median: 5
 Maximum: 9
 Mean: 4.5
 Std Deviation: 2.8779077367380057
Distribution of 100000 Samples:
 0: 10.116%
 1: 9.995%
 2: 9.903%
 3: 9.993%
 4: 10.005%
 5: 9.976%
 6: 9.89%
 7: 9.874%
 8: 10.11%
 9: 10.138%

Output Analysis: random_range(10)
Typical Timing: 94 ± 8 ns
Statistics of 256 Samples:
 Minimum: 0
 Median: 4
 Maximum: 9
 Mean: 4.27734375
 Std Deviation: 2.7089601461685278
Distribution of 100000 Samples:
 0: 9.98%
 1: 10.058%
 2: 9.94%
 3: 9.975%
 4: 10.018%
 5: 9.902%
 6: 9.836%
 7: 10.042%
 8: 10.104%
 9: 10.145%

Output Analysis: random_below(-10)
Typical Timing: 94 ± 8 ns
Statistics of 256 Samples:
 Minimum: -9
 Median: -5
 Maximum: 0
 Mean: -4.69921875
 Std Deviation: 2.933137143895928
Distribution of 100000 Samples:
 -9: 9.898%
 -8: 10.014%
 -7: 10.066%
 -6: 10.036%
 -5: 9.939%
 -4: 9.924%
 -3: 10.169%
 -2: 10.049%
 -1: 9.937%
 0: 9.968%

Output Analysis: random_index(-10)
Typical Timing: 63 ± 14 ns
Statistics of 256 Samples:
 Minimum: -10
 Median: -5
 Maximum: -1
 Mean: -5.390625
 Std Deviation: 2.756053942384534
Distribution of 100000 Samples:
 -10: 9.989%
 -9: 9.956%
 -8: 9.923%
 -7: 10.051%
 -6: 9.944%
 -5: 9.941%
 -4: 10.164%
 -3: 10.02%
 -2: 10.047%
 -1: 9.965%

Output Analysis: random_range(-10)
Typical Timing: 94 ± 10 ns
Statistics of 256 Samples:
 Minimum: -10
 Median: -6
 Maximum: -1
 Mean: -5.69140625
 Std Deviation: 2.8440657818638138
Distribution of 100000 Samples:
 -10: 9.996%
 -9: 10.253%
 -8: 9.999%
 -7: 10.005%
 -6: 9.945%
 -5: 9.931%
 -4: 10.009%
 -3: 9.922%
 -2: 9.989%
 -1: 9.951%

Base Case
Output Analysis: Random.randrange(1, 10)
Typical Timing: 1188 ± 31 ns
Statistics of 256 Samples:
 Minimum: 1
 Median: 5
 Maximum: 9
 Mean: 4.97265625
 Std Deviation: 2.600760517542483
Distribution of 100000 Samples:
 1: 10.984%
 2: 11.192%
 3: 11.2%
 4: 10.929%
 5: 11.2%
 6: 11.309%
 7: 10.938%
 8: 11.077%
 9: 11.171%

Output Analysis: random_range(1, 10)
Typical Timing: 94 ± 11 ns
Statistics of 256 Samples:
 Minimum: 1
 Median: 6
 Maximum: 9
 Mean: 5.23046875
 Std Deviation: 2.744979262573906
Distribution of 100000 Samples:
 1: 11.019%
 2: 11.298%
 3: 11.076%
 4: 11.068%
 5: 11.212%
 6: 11.164%
 7: 11.073%
 8: 11.056%
 9: 11.034%

Output Analysis: random_range(10, 1)
Typical Timing: 63 ± 12 ns
Statistics of 256 Samples:
 Minimum: 1
 Median: 5
 Maximum: 9
 Mean: 5.11328125
 Std Deviation: 2.614975607241985
Distribution of 100000 Samples:
 1: 10.992%
 2: 11.018%
 3: 11.269%
 4: 11.089%
 5: 11.067%
 6: 11.178%
 7: 11.103%
 8: 11.057%
 9: 11.227%

Base Case
Output Analysis: Random.randint(-5, 5)
Typical Timing: 1188 ± 21 ns
Statistics of 256 Samples:
 Minimum: -5
 Median: 0
 Maximum: 5
 Mean: -0.01171875
 Std Deviation: 3.056952759132635
Distribution of 100000 Samples:
 -5: 9.068%
 -4: 9.055%
 -3: 9.162%
 -2: 9.097%
 -1: 9.043%
 0: 9.111%
 1: 9.052%
 2: 8.993%
 3: 9.211%
 4: 9.043%
 5: 9.165%

Output Analysis: random_int(-5, 5)
Typical Timing: 63 ± 8 ns
Statistics of 256 Samples:
 Minimum: -5
 Median: 0
 Maximum: 5
 Mean: -0.27734375
 Std Deviation: 3.131940821411466
Distribution of 100000 Samples:
 -5: 9.25%
 -4: 9.034%
 -3: 9.098%
 -2: 9.163%
 -1: 9.028%
 0: 9.069%
 1: 8.973%
 2: 8.95%
 3: 9.16%
 4: 9.218%
 5: 9.057%

Base Case
Output Analysis: Random.randrange(1, 20, 2)
Typical Timing: 1250 ± 23 ns
Statistics of 256 Samples:
 Minimum: 1
 Median: 11
 Maximum: 19
 Mean: 9.734375
 Std Deviation: 5.7768766579847135
Distribution of 100000 Samples:
 1: 10.067%
 3: 10.125%
 5: 9.963%
 7: 10.14%
 9: 10.011%
 11: 9.912%
 13: 10.033%
 15: 9.944%
 17: 9.917%
 19: 9.888%

Output Analysis: random_range(1, 20, 2)
Typical Timing: 94 ± 12 ns
Statistics of 256 Samples:
 Minimum: 1
 Median: 9
 Maximum: 19
 Mean: 9.8515625
 Std Deviation: 5.556963261195318
Distribution of 100000 Samples:
 1: 10.067%
 3: 10.182%
 5: 9.925%
 7: 10.163%
 9: 9.856%
 11: 9.98%
 13: 9.804%
 15: 10.018%
 17: 10.213%
 19: 9.792%

Output Analysis: random_range(1, 20, -2)
Typical Timing: 94 ± 1 ns
Statistics of 256 Samples:
 Minimum: 2
 Median: 12
 Maximum: 20
 Mean: 10.84375
 Std Deviation: 5.58771411398327
Distribution of 100000 Samples:
 2: 9.967%
 4: 10.132%
 6: 10.07%
 8: 10.018%
 10: 9.949%
 12: 9.936%
 14: 10.028%
 16: 10.09%
 18: 9.825%
 20: 9.985%

Output Analysis: d(10)
Typical Timing: 63 ± 1 ns
Statistics of 256 Samples:
 Minimum: 1
 Median: 5
 Maximum: 10
 Mean: 5.234375
 Std Deviation: 2.8805891962664614
Distribution of 100000 Samples:
 1: 10.031%
 2: 10.053%
 3: 9.76%
 4: 10.064%
 5: 10.149%
 6: 9.95%
 7: 10.016%
 8: 10.035%
 9: 9.999%
 10: 9.943%

Output Analysis: dice(3, 6)
Typical Timing: 94 ± 16 ns
Statistics of 256 Samples:
 Minimum: 3
 Median: 11
 Maximum: 17
 Mean: 10.28125
 Std Deviation: 2.9196763183479035
Distribution of 100000 Samples:
 3: 0.47%
 4: 1.403%
 5: 2.704%
 6: 4.514%
 7: 6.663%
 8: 9.844%
 9: 11.71%
 10: 12.604%
 11: 12.473%
 12: 11.786%
 13: 9.6%
 14: 6.942%
 15: 4.648%
 16: 2.827%
 17: 1.339%
 18: 0.473%

Output Analysis: ability_dice(4)
Typical Timing: 219 ± 6 ns
Statistics of 256 Samples:
 Minimum: 4
 Median: 12
 Maximum: 18
 Mean: 12.20703125
 Std Deviation: 2.8587857999507413
Distribution of 100000 Samples:
 3: 0.076%
 4: 0.324%
 5: 0.771%
 6: 1.59%
 7: 2.937%
 8: 4.701%
 9: 6.95%
 10: 9.373%
 11: 11.471%
 12: 12.907%
 13: 13.223%
 14: 12.502%
 15: 10.138%
 16: 7.253%
 17: 4.169%
 18: 1.615%

Output Analysis: plus_or_minus(5)
Typical Timing: 63 ± 1 ns
Statistics of 256 Samples:
 Minimum: -5
 Median: 0
 Maximum: 5
 Mean: 0.17578125
 Std Deviation: 3.191342359246587
Distribution of 100000 Samples:
 -5: 9.041%
 -4: 8.942%
 -3: 9.114%
 -2: 9.072%
 -1: 9.338%
 0: 9.056%
 1: 9.161%
 2: 9.053%
 3: 9.063%
 4: 9.067%
 5: 9.093%

Output Analysis: plus_or_minus_linear(5)
Typical Timing: 63 ± 15 ns
Statistics of 256 Samples:
 Minimum: -5
 Median: 0
 Maximum: 5
 Mean: 0.1015625
 Std Deviation: 2.4409572013502263
Distribution of 100000 Samples:
 -5: 2.805%
 -4: 5.484%
 -3: 8.442%
 -2: 11.094%
 -1: 13.884%
 0: 16.766%
 1: 13.871%
 2: 11.166%
 3: 8.252%
 4: 5.439%
 5: 2.797%

Output Analysis: plus_or_minus_gauss(5)
Typical Timing: 125 ± 6 ns
Statistics of 256 Samples:
 Minimum: -4
 Median: 0
 Maximum: 4
 Mean: -0.01953125
 Std Deviation: 1.7499606088003643
Distribution of 100000 Samples:
 -5: 0.191%
 -4: 1.13%
 -3: 4.439%
 -2: 11.449%
 -1: 20.703%
 0: 24.692%
 1: 20.276%
 2: 11.33%
 3: 4.389%
 4: 1.188%
 5: 0.213%


Random Floats:

Base Case
Output Analysis: Random.random()
Typical Timing: 32 ± 16 ns
Statistics of 256 Samples:
 Minimum: 0.0028626903588546693
 Median: (0.4859747727918289, 0.49073531878526633)
 Maximum: 0.9871366303552271
 Mean: 0.49707105037008315
 Std Deviation: 0.2883104652819201
Post-processor Distribution of 100000 Samples using round():
 0: 50.116%
 1: 49.884%

Output Analysis: canonical()
Typical Timing: 32 ± 16 ns
Statistics of 256 Samples:
 Minimum: 0.003708474878451653
 Median: (0.4915840613555649, 0.492149623133669)
 Maximum: 0.9982321410679722
 Mean: 0.49375077574045045
 Std Deviation: 0.296387450899638
Post-processor Distribution of 100000 Samples using round():
 0: 50.03%
 1: 49.97%

Output Analysis: random_float(0.0, 10.0)
Typical Timing: 32 ± 16 ns
Statistics of 256 Samples:
 Minimum: 0.0309521847110069
 Median: (5.084236413548418, 5.115748263088992)
 Maximum: 9.972112689723225
 Mean: 5.216066481880628
 Std Deviation: 2.9029824200327923
Post-processor Distribution of 100000 Samples using floor():
 0: 9.898%
 1: 9.907%
 2: 10.036%
 3: 9.797%
 4: 10.094%
 5: 10.08%
 6: 10.122%
 7: 10.083%
 8: 9.931%
 9: 10.052%


Random Booleans:

Output Analysis: percent_true(33.33)
Typical Timing: 63 ± 1 ns
Statistics of 256 Samples:
 Minimum: False
 Median: False
 Maximum: True
 Mean: 0.33984375
 Std Deviation: 0.474583795985245
Distribution of 100000 Samples:
 False: 66.564%
 True: 33.436%


Shuffle Performance:

some_small_list = [i for i in range(10)]
some_med_list = [i for i in range(100)]
some_large_list = [i for i in range(1000)]

Base Case:
Random.shuffle()
Typical Timing: 8375 ± 146 ns
Typical Timing: 78375 ± 1171 ns
Typical Timing: 695125 ± 4121 ns

Test Case 1:
Fortuna.fisher_yates()
Typical Timing: 875 ± 58 ns
Typical Timing: 6375 ± 58 ns
Typical Timing: 75250 ± 149 ns

Test Case 2:
Fortuna.knuth()
Typical Timing: 875 ± 1 ns
Typical Timing: 6125 ± 65 ns
Typical Timing: 74250 ± 161 ns

Test Case 3:
Fortuna.shuffle() knuth_b algorithm
Typical Timing: 250 ± 45 ns
Typical Timing: 3000 ± 58 ns
Typical Timing: 29375 ± 115 ns


-------------------------------------------------------------------------
Total Test Time: 4.091 seconds

```


## Legal Information
Fortuna © 2019 Broken aka Robert W Sharp, all rights reserved.

Fortuna is licensed under a Creative Commons Attribution-NonCommercial 3.0 Unported License.

See online version of this license here: <http://creativecommons.org/licenses/by-nc/3.0/>
